/**
 * Introduction
 *
 * @title ROV - Easy applications from C
 *
 * @copyright Jason Tribbeck / 7th ARM 1994-2001
 *
 * @authors Jason Tribbeck
 *
 * @file genheap.c
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

typedef struct tags {
  int free:1;
  int len:31;
} tag;

#ifdef RISCOS
#define PTRLT(aa,bb) ((int)(aa)<(int)(bb))
#else
#define PTRLT(aa,bb) (aa<bb)
#endif

#ifdef VERBOSE
static void genheap_display(void * gmem, int len)
{
  tag * t = gmemory;

  len = len / sizeof(tag);
  tom = t + len;

  while(PTRLT(aa, bb)) {
    printf("Tag @ %p : ", t);
    if(t -> free) printf("free"); else printf("used");
    printf(" %d bytes\n", t -> len * sizeof(tag));
    t += t -> len;
  }
  printf("---------------------\n");
} 
#else
#define genheap_display(aa,bb)
#endif

void genheap_init(void * gmem, int len)
{
  register tag * gmemory = (tag *)gmem;
  register int memtags = len / sizeof(tag);

  gmemory[0].free = 1;
  gmemory[0].len = memtags;
  genheap_display(gmem, len);
}

void * genheap_malloc(void * gmem, int len, int size)
{
  tag * t = (tag *)gmem;
  tag * tom;

  len = len / sizeof(tag);
  tom = t + len;

  size = (size + sizeof(tag) - 1) / sizeof(tag);

  do {
    if((t -> free) && (size < t -> len)) {
      tag * t2;

      if(t -> len - size - 1) {
	t2 = t + size + 1;
	t2 -> len = t -> len - size - 1;
	t2 -> free = 1;
      }

      t -> len = size + 1;
      t -> free = 0;

      genheap_display(gmem, len);
      return (void *)(t + 1);
    }
    t += t -> len;
  } while(PTRLT(t, tom));
  return 0;
}

static void genheap_restructure(void * gmem, int len)
{
  tag * t = (tag *)gmem;
  tag * l = 0;
  tag * tom;

  len = len / sizeof(tag);
  tom = t + len;

  while(PTRLT(t, tom)) {
    if((l) && (l -> free) && (t -> free)) {
      l -> len += t -> len;
    } else {
      l = t;
    }
    t += t -> len;
  }
}

void genheap_resize(void * gmem, int oldlen, int newlen)
{
  tag * t = (tag *)gmem;
  tag * l = 0;
  tag * tom;

  oldlen = oldlen / sizeof(tag);
  newlen = newlen / sizeof(tag);
  tom = t + oldlen;

  while(PTRLT(t, tom)) {
    l = t;
    t += t -> len;
  }
  if(l -> free) {
    l -> len += newlen - oldlen;
  } else {
    tom -> free = 1;
    tom -> len = newlen - oldlen;
  }
}

void genheap_free(void * gmem, int len, void * ptr)
{
  tag * t = ((tag *)ptr) - 1;

  t -> free = 1;

  genheap_display(gmem, len);
  genheap_restructure(gmem, len);
  genheap_display(gmem, len);
}

int genheap_freespace(void * gmem, int len)
{
  int m = 0;
  tag * t = (tag *)gmem;
  tag * tom;

  len = len / sizeof(tag);
  tom = t + len;

  while(PTRLT(t, tom)) {
    if(t -> free) {
      m += (t -> len - 1) * sizeof(tag);
    }
    t += t -> len;
  }
  return m;
}

void * genheap_walk(void * gmem, int len, void * ptr)
{
  tag * tom;
  tag * t = (tag *)ptr - 1;

  len = len / sizeof(tag);
  tom = (tag *)gmem + len;

  if(ptr == NULL) {
    t = (tag *)gmem;
    return (void *)(t + 1);
  }
  t += t -> len;
  if(PTRLT(t, tom)) return (void *)(t + 1);
  return 0;
}

int genheap_size(void * ptr)
{
  tag * t = (tag *)ptr - 1;
  return t -> len * sizeof(tag);
}

char genheap_type(void * ptr)
{
  tag * t = (tag *)ptr - 1;
  if(t -> free) return 'f';
  return 'a';
}
