/**
 * Introduction
 *
 * @title ROV - Easy applications from C
 *
 * @copyright Jason Tribbeck / 7th ARM 1994-2001
 *
 * @authors Jason Tribbeck
 *
 * @file zap.c
 */

#include <stdio.h>
#include <stdarg.h>
#include <string.h>

#include "kernel.h"

#ifndef __GNUC__
#pragma -s1
#endif

#include "main.h"
#include "canvas.h"
#include "icons.h"
#include "file.h"
#include "msg.h"
#include "draw.h"

typedef struct {
  int flags;
  int x0, y0, x1, y1;
  int screen_address;
  int bytes_per_raster;
  int bits_per_pixel;
  int width, height;
  char * char_cache;
  int bytes_per_charline;
  int bytes_per_charcache;
  int line_spacing;
  char * rdata;
  int scx, scy;
  int * palette;
  int foreground, background;
  char * workarea;
  int shiftx, shifty;
  int screen_xpix, screen_ypix;
  int mode;
} zap_block;

static char * __zs;
static char * __bm;
static zap_block __zb;
static _kernel_swi_regs __zr;
static int __zp[0x40];
static char __zw[256];

void zap_modechange(void)
{
  int i;
  __zr.r[0] = 135;
  _kernel_swi(0x20006, & __zr, & __zr);
  if(__zb.mode == __zr.r[2]) return;

  __zr.r[1] = (int) & __zb;
  _kernel_swi(0x4848e, & __zr, & __zr);

  __zb.mode = __zr.r[2];
  if(draw_scaley == 2) {
    __zb.flags = 2;
  } else {
    __zb.flags = 0;
  }
  __zb.width = __zb.height = 8;
  __zb.char_cache = __bm;
  __zr.r[0] = __zb.bits_per_pixel;
  __zr.r[1] = 0;
  __zr.r[2] = __zr.r[3] = 8;
  _kernel_swi(0x48488, & __zr, & __zr);
  __zb.bytes_per_charline = __zr.r[2];
  __zb.bytes_per_charcache = __zr.r[3];
  __zb.line_spacing = 0;
  __zr.r[1] = (int) & __zb;
  __zr.r[2] = 32;
  __zr.r[3] = 255;
  __zr.r[4] = (int)__zs;
  _kernel_swi(0x48483, & __zr, & __zr);
  __zb.palette = __zp;
  for(i = 0; i < 16; i ++) {
    __zp[i] = i;
  }
  __zr.r[0] = 3;
  __zr.r[1] = (int) & __zb;
  __zr.r[2] = __zr.r[3] = (int)__zp;
  __zr.r[4] = 16;
  _kernel_swi(0x4848a, & __zr, & __zr);
  __zb.background = 0;
  __zb.foreground = 7;
}

void zap_init(void)
{
  __zs = rmalloc(0x800);
  __bm = rmalloc(0x10000);
  memset(__zs, 0xff, 0x800);
  memset(__bm, 0xff, 0x10000);
  __zb.workarea = __zw;
  __zb.rdata = 0;
  __zr.r[0] = (int)__zs;
  __zr.r[1] = (int) & __zb;
  _kernel_swi(0x4848c, & __zr, & __zr);
  __zb.mode = -1;
  zap_modechange();
}

void zap_redraw(int * lines, int fore, int back, canvas c)
{
  memset(& __zr, 0, sizeof(_kernel_swi_regs));
  if(canvas_redrawmode == 0) {
    __zr.r[0] = (int) & poll_block;
  } else {
    __zr.r[0] = (int)canvas_windowptr(c);
  }
  __zr.r[1] = (int) & __zb;
  __zb.foreground = fore;
  __zb.background = back;
  __zb.rdata = (char *)lines;
  _kernel_swi(0x4848f, & __zr, & __zr);
  _kernel_swi(0x48480, & __zr, & __zr);
}

int * zap_processtext(char * text, int * l)
{
  char * a = text;
  int lines = 0, * i;

  while((a[0]) || (a[1]) || (a[2])) {
    if((* a == 0) && (a[1] == 2)) lines ++ ;
    a ++ ;
  }
  if(l == 0) l = rmalloc(lines * 4 + 4);
  a = text;
  i = l;
  * i ++ = (int)a - (int)l;
  while((a[0]) || (a[1]) || (a[2])) {
    if((* a == 0) && (a[1] == 2)) * i ++ = (int)a - (int)l + 2;
    a ++ ;
  }
  i[ - 1] = 0;
  __zb.rdata = (char *)l;
  return l;
}
